from base import EntityDescriptor, UniqueName


class AnalogAlarmDescriptor:
    """Board Descriptor of analog alarms"""

    _enum_indexes = list()

    def __init__(self, enum_index, enable_default, threshold_default, timeout_default_ms):
        """
        Constructor
        :param enum_index: index used if used if enumerating
        :param enable_default: boolean for default alarm enable
        :param threshold_default: threshold value expressed as a float value (physical value)
        :param timeout_default_ms: default timeout value in milliseconds
        """
        self.enum_index = enum_index
        self.enable_default = enable_default
        self.threshold_default = threshold_default
        self.timeout_default_ms = timeout_default_ms
        # check enum uniqueness
        if self.enum_index is not None:
            assert self.enum_index not in self._enum_indexes, 'Enum index {} already used for class "{}"'.format(self.enum_index, self.__class__.__name__)
        self._enum_indexes.append(self.enum_index)


class AnalogVariableDescriptor(EntityDescriptor):
    """Board Analog variable descriptor"""

    _enum_indexes = list()

    def __init__(self, firmware_config, name, analog_channel, channel2measure_converter, measure2channel_converter, units, description, enum_index=None, alarm_slow=None, alarm_fast=None):
        """
        Constructor
        :param firmware_config: the class FirmwareConfig
        :param name: name of the analog variable (used for code generation)
        :param analog_channel: ADC channel number
        :param channel2measure_converter: function to convert from the ADC number (unsigned) to float value (physical value)
        :param measure2channel_converter: function to convert from float value (physical value) to the ADC number (unsigned)
        :param description: human readeable description (used for comments)
        :param enum_index: index used if used if enumerating
        :param alarm_slow: default values for the slow alarm
        :param alarm_fast: default values for the slow alarm
        """
        EntityDescriptor.__init__(self, firmware_config, name)
        self.analog_channel = analog_channel
        self.channel2measure_converter = channel2measure_converter
        self.measure2channel_converter = measure2channel_converter
        self.units = units
        self.description = description
        self.enum_index = enum_index
        self.alarm_slow = alarm_slow
        self.alarm_fast = alarm_fast
        # check enum uniqueness
        if self.enum_index is not None:
            assert self.enum_index not in self._enum_indexes, 'Enum index {} already used for class "{}"'.format(self.enum_index, self.__class__.__name__)
        self._enum_indexes.append(self.enum_index)


    def get_alarm_callback_name(self):
        return 'alarm_' + self.name

    def get_alarm_fast_timeout_ticks(self, num_analogvariables):
        # calculate the number of ticks of the timeout
        timeout_ticks = round((self.alarm_fast.timeout_default_ms / self.firmware_config.SLOW_TIMER_PERIOD_MS) / (self.firmware_config.ADCREADER_CALLS_PER_MEASURE * num_analogvariables))
        assert timeout_ticks < self.firmware_config.TIMEOUT_COUNTER_MAX, 'get_alarm_fast_timeout_ticks: {}'.format(timeout_ticks)
        return timeout_ticks

    def get_alarm_slow_timeout_ticks(self, num_analogvariables):
        # calculate the number of ticks of the timeout
        timeout_ticks = round((self.alarm_slow.timeout_default_ms / self.firmware_config.SLOW_TIMER_PERIOD_MS) / (self.firmware_config.ADCREADER_CALLS_PER_MEASURE * num_analogvariables))
        assert timeout_ticks < self.firmware_config.TIMEOUT_COUNTER_MAX, 'get_alarm_slow_timeout_ticks: {}'.format(timeout_ticks)
        return timeout_ticks

    def get_alarm_fast_timeout_actual_ms(self, num_analogvariables):
        # calculate the actual number of milliseconds of the timeout
        return self.get_alarm_fast_timeout_ticks(num_analogvariables) * self.firmware_config.SLOW_TIMER_PERIOD_MS * (self.firmware_config.ADCREADER_CALLS_PER_MEASURE * num_analogvariables)

    def get_alarm_slow_timeout_actual_ms(self, num_analogvariables):
        # calculate the actual number of milliseconds of the timeout
        return self.get_alarm_slow_timeout_ticks(num_analogvariables) * self.firmware_config.SLOW_TIMER_PERIOD_MS * (self.firmware_config.ADCREADER_CALLS_PER_MEASURE * num_analogvariables)

    def get_alarm_fast_threshold_channels(self):
        threshold_channels = round(self.measure2channel_converter(self.alarm_fast.threshold_default))
        assert threshold_channels < self.firmware_config.ADC_NUM_CHANNELS, 'get_alarm_fast_threshold_channels too high {} on {}'.format(threshold_channels, self.name)
        return threshold_channels

    def get_alarm_slow_threshold_channels(self):
        threshold_channels = round(self.measure2channel_converter(self.alarm_slow.threshold_default))
        assert threshold_channels < self.firmware_config.ADC_NUM_CHANNELS, 'get_alarm_slow_threshold_channels too high {} on {}'.format(threshold_channels, self.name)
        return threshold_channels


class DigitalAlarmDescriptor:
    """Board Descriptor of digital alarms"""

    def __init__(self, enum_index, enable_default, timeout_default_ms):
        """
        Constructor
        :param enum_index: index used if used if enumerating
        :param enable_default: boolean for default alarm enable
        :param timeout_default_ms: default timeout value in milliseconds
        """
        self.enum_index = enum_index
        self.enable_default = enable_default
        self.timeout_default_ms = timeout_default_ms


class DigitalVariableDescriptor(EntityDescriptor):
    """Board Digital variable descriptor"""

    _enum_indexes = list()

    def __init__(self, firmware_config, name, ioname, description, enum_index=None, alarm=None):
        """
        Coonbstructor
        :param firmware_config: the class FirmwareConfig
        :param name: name of the digital variable (used for code generation)
        :param ioname: name of the pin control macro (i.e. _SetHigh and _SetLow will be appended)
        :param description: human readeable description (used for comments)
        :param enum_index: index used if used if enumerating
        :param alarm: default values for the alarm
        """
        EntityDescriptor.__init__(self, firmware_config, name)
        self.ioname = ioname
        self.description = description
        self.enum_index = enum_index
        self.alarm = alarm
        # check enum uniqueness
        if self.enum_index is not None:
            assert self.enum_index not in self._enum_indexes, 'Enum index {} already used for class "{}"'.format(self.enum_index, self.__class__.__name__)
        self._enum_indexes.append(self.enum_index)

    def get_alarm_callback_name(self):
        return 'alarm_' + self.name

    def get_read_function_name(self):
        return 'ioread_' + self.ioname

    def get_alarm_timeout_ticks(self):
        # calculate the number of ticks of the timeout
        timeout_ticks = round(self.alarm.timeout_default_ms / self.firmware_config.FAST_TIMER_PERIOD_MS)
        assert timeout_ticks < self.firmware_config.TIMEOUT_COUNTER_MAX, 'get_alarm_timeout_ticks: {}'.format(timeout_ticks)
        return timeout_ticks

    def get_alarm_timeout_actual_ms(self):
        # calculate the actual number of milliseconds of the timeout
        return self.get_alarm_timeout_ticks() * self.firmware_config.FAST_TIMER_PERIOD_MS


class SwitchDescriptor(EntityDescriptor):
    """Board Switch descriptor"""

    _enum_indexes = list()

    def __init__(self, firmware_config, name, ioname, description, enum_index=None):
        """
        Constructor
        :param firmware_config: the class FirmwareConfig
        :param name: name of the switch (used for code generation)
        :param ioname: name of the pin control macro (i.e. _SetHigh and _SetLow will be appended)
        :param description: human readeable description (used for comments)
        :param enum_index: index used if used if enumerating
        """
        EntityDescriptor.__init__(self, firmware_config, name)
        self.ioname = ioname
        self.description = description
        self.enum_index = enum_index
        # check enum uniqueness
        if self.enum_index is not None:
            assert self.enum_index not in self._enum_indexes, 'Enum index {} already used for class "{}"'.format(self.enum_index, self.__class__.__name__)
            self._enum_indexes.append(self.enum_index)


class UserPinDescriptor(EntityDescriptor):
    """Board user pin descriptor"""

    _enum_indexes = list()

    def __init__(self, firmware_config, name, ioname, description, enum_index=None):
        """
        Constructor
        :param firmware_config: the class FirmwareConfig
        :param name: name of the pin (used for code generation)
        :param ioname: name of the pin control macro (i.e. _SetHigh and _SetLow will be appended)
        :param description: human readeable description (used for comments)
        :param enum_index: index used if used if enumerating
        """
        EntityDescriptor.__init__(self, firmware_config, name)
        self.ioname = ioname
        self.description = description
        self.enum_index = enum_index
        # check enum uniqueness
        if self.enum_index is not None:
            assert self.enum_index not in self._enum_indexes, 'Enum index {} already used for class "{}"'.format(self.enum_index, self.__class__.__name__)
            self._enum_indexes.append(self.enum_index)


class PayloadFieldDescriptor:
    """A single payload (request or response) field in a command"""

    def __init__(self, name, ctype, description, length):
        """
        Constructor
        :param name:
        :param ctype:
        :param description:
        """
        self.name = name
        self.ctype = ctype
        self.description = description
        self.length = length


class PayloadFieldU8(PayloadFieldDescriptor):
    """A single payload field containing an uint8"""

    def __init__(self, name, description):
        # super's constructor
        PayloadFieldDescriptor.__init__(self, name, 'uint8_t', description, length=1)


class PayloadFieldU16(PayloadFieldDescriptor):
    """A single payload field containing an uint16"""

    def __init__(self, name, description):
        # super's constructor
        PayloadFieldDescriptor.__init__(self, name, 'uint16_t', description, length=2)


class PayloadFieldU32(PayloadFieldDescriptor):
    """A single payload field containing an uint32"""

    def __init__(self, name, description):
        # super's constructor
        PayloadFieldDescriptor.__init__(self, name, 'uint32_t', description, length=4)


class PayloadFieldEnum(PayloadFieldDescriptor):
    """A single payload field containing an enum value (mapped to an u8)"""

    _persistent_list = {}
    _identifiers = []

    def __init__(self, name, description, index_list, value_list):
        """
        :param name: string used in C code to identify the payload field
        :param description: description of the payload field
        :param index_list: ordered list of indexes of the enum field
        :param value_list: ordered list of values associated to the given enums (usually strings)
        """

        PayloadFieldDescriptor.__init__(self, name, 'uint8_t', description, length=1)
        assert len(index_list) == len(value_list)
        assert max(index_list) < 255  # because we map to uint8_t
        self.index_list = index_list
        self.value_list = value_list
        # self.reference_table_num = len(PayloadFieldEnum.persistent_list)
        # compute long description field
        indexing_desc = '\n'.join([(str(index_list[k]) + ' -> ' + value_list[k]) for k in range(len(index_list))])
        self.detailed_description = 'Enum values:' + '\n{}'.format(indexing_desc)
        # verify the name is not already used (warn user)
        if self.name in PayloadFieldEnum._identifiers:
            print "WARNING: Identifier {} already used (now in class {})".format(self.name, self.__class__.__name__)
        else:
            PayloadFieldEnum._identifiers.append(self.name)
            PayloadFieldEnum._persistent_list[self.name] = self
        # assert self.name not in PayloadFieldEnum._identifiers, "Identifier {} already used.".format(self.name)

    def generate_enum_typename(self):
        return 'enum_' + self.name.lower() + '_t'

    def get_index_min(self):
        return min(self.index_list)

    def get_index_max(self):
        return max(self.index_list)

    def get_value_by_index(self, index):
        for idx in range(len(self.index_list)):
            if self.index_list[idx] == index:
                return self.value_list[idx]
        return None


    @staticmethod
    def get_persistent_list():
        return PayloadFieldEnum._persistent_list


class CommandDescriptor(UniqueName):
    """Board Command descriptor"""

    _command_codes = list()

    def __init__(self, firmware_config, name, request_code, response_code, description, request_payload, response_payload, implement=True):
        """
        Constructor
        :param firmware_config: the class FirmwareConfig
        :param name: command name (used for code generation)
        :param request_code: request code number (set to None for auto-numbering)
        :param response_code: response code number (set to None for auto-numbering)
        :param description: command description (for comments)
        :param request_payload: the request payload (an ordered list of PayloadField)
        :param response_payload: the response payload (an ordered list of PayloadField)
        :param implement: if False the command must be considered invalid (will not be callable by host)
        """
        # check request and response payload are correct type
        assert all(isinstance(x, PayloadFieldDescriptor) for x in request_payload), "All request payload fields must be of type PayloadField"
        assert all(isinstance(x, PayloadFieldDescriptor) for x in response_payload), "All response payload fields must be of type PayloadField"
        UniqueName.__init__(self, name)
        self.firmware_config = firmware_config
        self.request_code = request_code
        self.response_code = response_code
        self.description = description
        self.request_payload = request_payload
        self.response_payload = response_payload
        self.implement = implement
        # check command codes not already used
        if self.request_code is not None:
            assert self.request_code not in self._command_codes, 'Request command code {} already used for class "{}"'.format(self.request_code, self.__class__.__name__)
            self._command_codes.append(self.request_code)
        if self.response_code is not None:
            assert self.response_code not in self._command_codes, 'Response command code {} already used for class "{}"'.format(self.response_code, self.__class__.__name__)
            self._command_codes.append(self.response_code)

    def generate_callback_name(self):
        return 'command_' + self.name

    def get_request_payload_len(self):
        return sum([itm.length for itm in self.request_payload])

    def get_response_payload_len(self):
        return sum([itm.length for itm in self.response_payload])




