from analogvariables import AnalogVariableList
from entities.descriptors import PayloadFieldEnum


def generate_analogvariables_header(_analogvariables):

    includes = """
#include "../../typedefs.h"
"""

    typedef = """
// analog reading info structure
typedef struct {
    analog_channel_t channel;                   // ADC channel from which the variable can be read
    analog_value_t value;                       // last acquired value
    analog_variable_sum_t sum_curr;             // current sum (to compute average)
    analog_variable_sum_t sum_last;             // last computed sum
    analog_variable_sum_counter_t sum_counter;  // current count for sum computation
    analog_value_t offset;                      // offset value
    analog_value_t value_max;                   // max value acquired
    analogalarm_dual_info_t alarms;             // structure maintaining all the info related to the slow and fast alarms    
} analog_variable_t;
    """

    countdef = """
// number of defined analog variables
#define ANALOG_VARIABLES_COUNT {}
    """.format(_analogvariables.count())

    extern = """
    
// make analog variable list available to anyone including this file
extern analog_variable_t analog_variables[ANALOG_VARIABLES_COUNT];
    """

    varindex_tpl = """
#define ANALOG_VARIABLE_INDEX_[varname]  [index]    // index of the analog variable [varname] inside analog_variables array"""

    varindex_contents = ''
    index = 0
    for variable in _analogvariables.entries():
        varindex_define = varindex_tpl
        varindex_define = varindex_define.replace('[varname]', variable.name)
        varindex_define = varindex_define.replace('[index]', str(index))
        index += 1
        varindex_contents += varindex_define

    header_contents = includes + typedef + countdef + varindex_contents + extern

    return header_contents


def generate_analogvariables_source(_analogvariables):

    entry_tpl = """// Variable: [name] ([description])
{
    .channel = [channel],
    .value = 0,
    .sum_curr = 0,
    .sum_last = 0,
    .sum_counter = 0,
    .offset = 0,
    .value_max = 0,
    .alarms = {
        .fast = {
            .enabled = [fast_alarm_enabled],
            .threshold = [fast_alarm_threshold], // [fast_alarm_threshold_comment]
            .timeout = [fast_alarm_timeout], // [fast_alarm_timeout_comment]
            .timeout_counter = 0,
            .timeout_counter_max = 0,
            .firecount = 0
        },
        .slow = {
            .enabled = [slow_alarm_enabled],
            .threshold = [slow_alarm_threshold], // [slow_alarm_threshold_comment]
            .timeout = [slow_alarm_timeout], // [slow_alarm_timeout_comment]
            .timeout_counter = 0,
            .timeout_counter_max = 0,
            .firecount = 0
        },
        .callback = [callback]
    }
}"""

    source_list = list()
    for analogvariable in _analogvariables.entries():

        fast_alarm_timeout = analogvariable.get_alarm_fast_timeout_ticks(_analogvariables.count())
        slow_alarm_timeout = analogvariable.get_alarm_slow_timeout_ticks(_analogvariables.count())
        fast_alarm_timeout_ms = analogvariable.get_alarm_fast_timeout_actual_ms(_analogvariables.count())
        slow_alarm_timeout_ms = analogvariable.get_alarm_slow_timeout_actual_ms(_analogvariables.count())

        fast_alarm_threshold = analogvariable.get_alarm_fast_threshold_channels()
        slow_alarm_threshold = analogvariable.get_alarm_slow_threshold_channels()

        callback = analogvariable.get_alarm_callback_name()

        entry_source = entry_tpl
        entry_source = entry_source.replace('[name]', analogvariable.name)
        entry_source = entry_source.replace('[description]', analogvariable.description)
        entry_source = entry_source.replace('[channel]', str(analogvariable.analog_channel))
        entry_source = entry_source.replace('[fast_alarm_enabled]', 'true' if analogvariable.alarm_slow.enable_default else 'false')
        entry_source = entry_source.replace('[fast_alarm_threshold]', str(int(fast_alarm_threshold)))
        entry_source = entry_source.replace('[fast_alarm_threshold_comment]', str(analogvariable.alarm_fast.threshold_default) + ' ' + analogvariable.units)
        entry_source = entry_source.replace('[fast_alarm_timeout]', str(int(fast_alarm_timeout)))
        entry_source = entry_source.replace('[fast_alarm_timeout_comment]', '{} ms ({} ms real)'.format(analogvariable.alarm_fast.timeout_default_ms, fast_alarm_timeout_ms))
        entry_source = entry_source.replace('[slow_alarm_enabled]', 'true' if analogvariable.alarm_slow.enable_default else 'false')
        entry_source = entry_source.replace('[slow_alarm_threshold]', str(int(slow_alarm_threshold)))
        entry_source = entry_source.replace('[slow_alarm_threshold_comment]',  str(analogvariable.alarm_slow.threshold_default) + ' ' + analogvariable.units)
        entry_source = entry_source.replace('[slow_alarm_timeout]', str(int(slow_alarm_timeout)))
        entry_source = entry_source.replace('[slow_alarm_timeout_comment]', '{} ms ({} ms real)'.format(analogvariable.alarm_slow.timeout_default_ms, slow_alarm_timeout_ms))
        entry_source = entry_source.replace('[callback]', callback)

        source_list.append(entry_source)

    source_contents = ''
    source_contents += '#include "variables.h"\n'
    source_contents += '# include "../../configuration.h"\n\n'

    source_contents += 'analog_variable_t analog_variables[ANALOG_VARIABLES_COUNT] = {\n'
    source_contents += ','.join(source_list)
    source_contents += '};\n'

    return source_contents


def generate_analogalarm_callbacks_header(_analogvariables):
    alarm_callback_header_tpl = """
// Callback function for alarm on analog variable [name] ([description])
void [callback]();
"""
    source_list = list()
    for analogvariable in _analogvariables.entries():
        source_list.append(
            alarm_callback_header_tpl
                .replace('[name]', analogvariable.name)
                .replace('[description]', analogvariable.description)
                .replace('[callback]', analogvariable.get_alarm_callback_name())
        )
    return ''.join(source_list)


def generate_analogalarm_callbacks_source_template(_analogvariables):
    alarm_callback_source_tpl = """
// Callback function for alarm on analog variable [name] ([description])
void [callback]() {
    // TODO implement auto-generated callback function [callback]
}
"""
    source_list = list()
    for analogvariable in _analogvariables.entries():
        source_list.append(
            alarm_callback_source_tpl
                .replace('[name]', analogvariable.name)
                .replace('[description]', analogvariable.description)
                .replace('[callback]', analogvariable.get_alarm_callback_name())
        )

    contents = '#include "generated/sources/variables.h"\n'
    contents += ''.join(source_list)
    return contents


if __name__ == '__main__':
    analogvariables = AnalogVariableList()
    print generate_analogvariables_header(analogvariables)
    print generate_analogvariables_source(analogvariables)
    print generate_analogalarm_callbacks_header(analogvariables)
    print generate_analogalarm_callbacks_source_template(analogvariables)


